from pyfbsdk import *
from pyfbsdk_additions import *
from math import sqrt

lMgr = FBConstraintManager()

def CreateParentChildConstraint(pConstraintManager):
    for lIdx in range( pConstraintManager.TypeGetCount() ):
        if pConstraintManager.TypeGetName(lIdx) == 'Parent/Child':
            return pConstraintManager.TypeCreateConstraint(lIdx)
    return None

def EventContainerDblClick(control, event):
    tool.model = None
    tool.container.Items.removeAll()
    
def EventContainerDragAndDrop(control, event):
    if event.State == FBDragAndDropState.kFBDragAndDropDrag:
        if event.Components[0].Is(FBModel_TypeInfo()):
            event.Accept()
    elif event.State == FBDragAndDropState.kFBDragAndDropDrop:
        tool.model = event.Components[0]
        tool.container.Items.removeAll()
        tool.container.Items.append(tool.model.Name)

def EventContainerPRBDblClick(control, event):
    tool.parentRB = None
    tool.containerPRB.Items.removeAll()
    
def EventContainerPRBDragAndDrop(control, event):
    if event.State == FBDragAndDropState.kFBDragAndDropDrag:
        if event.Components[0].Is(FBModel_TypeInfo()):
            event.Accept()
    elif event.State == FBDragAndDropState.kFBDragAndDropDrop:
        tool.parentRB = event.Components[0]
        tool.containerPRB.Items.removeAll()
        tool.containerPRB.Items.append(tool.parentRB.Name)

def EventContainerRBDblClick(control, event):
    tool.rb = None
    tool.containerRB.Items.removeAll()
    
def EventContainerRBDragAndDrop(control, event):
    if event.State == FBDragAndDropState.kFBDragAndDropDrag:
        if event.Components[0].ClassName() == "KxL_RigidBodyProperty":
            event.Accept()
    elif event.State == FBDragAndDropState.kFBDragAndDropDrop:
        tool.rb = event.Components[0]
        tool.containerRB.Items.removeAll()
        tool.containerRB.Items.append(tool.rb.Name)
        
def EventContainerJDblClick(control, event):
    tool.joint = None
    tool.containerJ.Items.removeAll()
    
def EventContainerJDragAndDrop(control, event):
    if event.State == FBDragAndDropState.kFBDragAndDropDrag:
        if event.Components[0].ClassName() == "KxL_JointBasicProperty":
            event.Accept()
    elif event.State == FBDragAndDropState.kFBDragAndDropDrop:
        tool.joint = event.Components[0]
        tool.containerJ.Items.removeAll()
        tool.containerJ.Items.append(tool.joint.Name)        

def GetRoot(pModel):
    tmp = pModel
    while tmp.Parent:
        tmp = tmp.Parent
    return tmp

def GetLength(pV):
    return sqrt(pV[0]*pV[0]+pV[1]*pV[1]+pV[2]*pV[2])

def GetNormalized(pV):
    lLength = GetLength(pV)
    return FBVector3d(pV[0]/lLength,pV[1]/lLength,pV[2]/lLength)

def GetCross(pV1, pV2):
    return FBVector3d(pV1[1]*pV2[2] - pV1[2]*pV2[1], pV1[2]*pV2[0] - pV1[0]*pV2[2], pV1[0]*pV2[1] - pV1[1]*pV2[0])

def GetDirMatrix(pDir, pPos):
    lDir_X = FBVector3d()
    lDir_Y = FBVector3d()
    lDir_Z = GetNormalized(pDir)
    if(lDir_Z[2] > - 0.9999999 and lDir_Z[2] < 0.9999999):
        lDir_Y = GetNormalized(FBVector3d(lDir_Z[0]*lDir_Z[2],lDir_Z[1]*lDir_Z[2],(lDir_Z[2]*lDir_Z[2])+1))
        lDir_X = GetNormalized(GetCross(lDir_Y,lDir_Z))
        lDir_Y = GetNormalized(GetCross(lDir_Z,lDir_X))
    else:
        lDir_X = GetNormalized(FBVector3d(lDir_Z[2],0,-lDir_Z[0]))
        lDir_Y[1] = 1
    lM = FBMatrix()
    for i in range(0,3):
        lM[i] = lDir_X[i]
        lM[4+i] = lDir_Y[i]
        lM[8+i] = lDir_Z[i]
        lM[12+i] = pPos[i]
    return lM

def SetupJ_Connections(pJointModel, pRB1, pRB2):
    #lName = pJointModel.Name +" Joint.Connections"
    #lConnections = tool.joint.PropertyList.Find(lName,False)
    lConnections = None
    lName = pJointModel.Name +" Joint"
    for lComp in tool.joint.Components:
        if lComp.Name == lName:
            lConnections = lComp.PropertyList.Find("Connections",False)
    if lConnections != None:
        if pRB1:
            FBConnect(pRB1,lConnections)
        if pRB2:
            FBConnect(pRB2,lConnections)

def SetupRB_J_Between(pModel1, pModel2, pPreviouseRB):
    lV1 = FBVector3d()
    pModel1.GetVector(lV1)
    lMatrix = FBMatrix()
    lRB = None
    if pModel2:
        lV2 = FBVector3d()
        pModel2.GetVector(lV2)
        lDir = FBVector3d(lV2[0] - lV1[0],lV2[1] - lV1[1],lV2[2] - lV1[2])
        lCenter = FBVector3d(lV1[0] + lDir[0]/2, lV1[1] + lDir[1]/2, lV1[2] + lDir[2]/2)
        lLength = GetLength(lDir)*100
        lRadius = tool.radius.Value
        if tool.globalradius.State == 1:
            lRadius *= lLength
        else:
            lRadius *= 100
        #create RB model
        lRB = FBModelMarker("Chain RB")
        lRB.Look = FBMarkerLook.kFBMarkerLookCapsule
        lRB.Size = lRadius
        lRB.Length = lLength
        lMatrix = GetDirMatrix(lDir,lCenter)
        lRB.SetMatrix(lMatrix)
        lRB.Show = True
        FBConnect(tool.rb, lRB)
    else:
        pModel1.GetMatrix(lMatrix,FBModelTransformationMatrix.kModelRotation)
    if pPreviouseRB or (tool.pinroot.State == True or tool.parentRB):
        #create J model
        lJ = FBModelNull("Chain J")
        lJ.Size = 400
        lJ.SetVector(lV1)
        lJ.SetMatrix(lMatrix,FBModelTransformationMatrix.kModelRotation)
        lJ.Show = True
        FBConnect(tool.joint, lJ)
        if pPreviouseRB:
            SetupJ_Connections(lJ,pPreviouseRB,lRB)
        else:
            SetupJ_Connections(lJ,tool.parentRB,lRB)
    if tool.constrainskeleton.State == True:
        lCnst = CreateParentChildConstraint(lMgr)
        lCnst.ReferenceAdd(0,pModel1)
        lCnst.ReferenceAdd(1,lRB)
        #lCnst.Snap()
        #lCnst.Lock = True
        
    return lRB
    
def CreateRB_J(pParentModel, pPreviouseRB):
    if len(pParentModel.Children) > 0:
        for lChild in pParentModel.Children:
            lRB = SetupRB_J_Between(pParentModel, lChild, pPreviouseRB)
            CreateRB_J(lChild,lRB)
    elif tool.pinleafs.State == True:
        SetupRB_J_Between(pParentModel, None, pPreviouseRB)

def EventButtonCreate(control, event):
    if tool.model == None:
        print "No chain to convert! Please drop model into container."
    else:
        if tool.rb == None:
            tool.rb = FBCreateObject("Browsing/Templates/Physical Properties", "Rigid Body","Chain RB Property" )
            tool.containerRB.Items.append(tool.rb.Name)
        if tool.joint == None:
            tool.joint = FBCreateObject("Browsing/Templates/Physical Properties", "Joint","Chain Joint Property" )
            tool.containerJ.Items.append(tool.joint.Name)
        if tool.searchroot.State == True:
            CreateRB_J(GetRoot(tool.model),None)
        else:
            CreateRB_J(tool.model,None)

def PopulateLayout(mainLyt):    
    x = FBAddRegionParam(5,FBAttachType.kFBAttachLeft,"")
    y = FBAddRegionParam(-5,FBAttachType.kFBAttachTop,"")
    w = FBAddRegionParam(-5,FBAttachType.kFBAttachRight,"")
    h = FBAddRegionParam(25,FBAttachType.kFBAttachBottom,"")
    mainLyt.AddRegion("main","main", x, y, w, h)
    vlyt = FBVBoxLayout()
    mainLyt.SetControl("main",vlyt)
    
    l = FBLabel()
    l.Caption = "Drag chain model. Double click to clear."
    vlyt.Add(l,30)
    
    tool.model = None
    tool.container = FBContainer()
    tool.container.ItemWidth = 200
    tool.container.OnDragAndDrop.Add(EventContainerDragAndDrop)
    tool.container.OnDblClick.Add(EventContainerDblClick)
    vlyt.Add(tool.container,30)
    
    l = FBLabel()
    l.Caption = "Drag chain parent. Double click to clear."
    vlyt.Add(l,30)
    
    tool.parentRB = None
    tool.containerPRB = FBContainer()
    tool.containerPRB.ItemWidth = 200
    tool.containerPRB.OnDragAndDrop.Add(EventContainerPRBDragAndDrop)
    tool.containerPRB.OnDblClick.Add(EventContainerPRBDblClick)
    vlyt.Add(tool.containerPRB,30)
    
    lRB = FBLabel()
    lRB.Caption = "Drag rigid body property. Double click to clear."
    vlyt.Add(lRB,30)
    
    tool.rb = None
    tool.containerRB = FBContainer()
    tool.containerRB.ItemWidth = 200
    tool.containerRB.OnDragAndDrop.Add(EventContainerRBDragAndDrop)
    tool.containerRB.OnDblClick.Add(EventContainerRBDblClick)
    vlyt.Add(tool.containerRB,30)
    
    lJ = FBLabel()
    lJ.Caption = "Drag joint property. Double click to clear."
    vlyt.Add(lJ,30)
    
    tool.joint = None
    tool.containerJ = FBContainer()
    tool.containerJ.ItemWidth = 200
    tool.containerJ.OnDragAndDrop.Add(EventContainerJDragAndDrop)
    tool.containerJ.OnDblClick.Add(EventContainerJDblClick)
    vlyt.Add(tool.containerJ,30)
    
    tool.searchroot = FBButton()
    tool.searchroot.Caption = "Look for root of hierarchy"
    tool.searchroot.Style = FBButtonStyle.kFBCheckbox 
    vlyt.Add(tool.searchroot,20)
    tool.pinroot = FBButton()
    tool.pinroot.Caption = "Pin Root model to World"
    tool.pinroot.Style = FBButtonStyle.kFBCheckbox 
    vlyt.Add(tool.pinroot,20)
    tool.pinleafs = FBButton()
    tool.pinleafs.Caption = "Pin All Leafs models to World"
    tool.pinleafs.Style = FBButtonStyle.kFBCheckbox 
    vlyt.Add(tool.pinleafs,20)
    tool.constrainskeleton = FBButton()
    tool.constrainskeleton.Caption = "Attach Parent/Child contraint to Skeleton Nodes"
    tool.constrainskeleton.Style = FBButtonStyle.kFBCheckbox 
    vlyt.Add(tool.constrainskeleton,20)
    
    hlyt = FBHBoxLayout(FBAttachType.kFBAttachLeft)
    label = FBLabel()
    label.Caption = "Radius: "
    hlyt.Add(label,40)
    tool.radius = FBEditNumber()
    tool.radius.Min = 0
    tool.radius.Max = 1000
    tool.radius.Value = 0.1
    hlyt.Add(tool.radius,60)
    tool.globalradius = FBButton()
    tool.globalradius.Caption = "Length Dependent"
    tool.globalradius.State = 1
    tool.globalradius.Style = FBButtonStyle.kFBCheckbox 
    hlyt.Add(tool.globalradius,140)
    vlyt.Add(hlyt,20)
    
    tool.button = FBButton()
    tool.button.Caption = "Create"
    tool.button.OnClick.Add(EventButtonCreate)
    vlyt.Add(tool.button,50)
    

def CreateTool():
    global tool
    
    tool = FBCreateUniqueTool("Joint Tool")
    tool.StartSizeX = 250
    tool.StartSizeY = 470
    PopulateLayout(tool)
    ShowTool(tool)
    

CreateTool()